import enum
import logging
import platform
import subprocess
import sys
from typing import Optional

from klisch import logger

log = logger.get(__file__, logging.DEBUG)


class Platform(enum.Enum):
    WINDOWS = 'Windows'
    LINUX = 'Linux'
    WSL = 'Windows Subsystem for Linux'
    LSF = 'Platform Load Sharing Facility'


class Shell(enum.Enum):
    TCSH = 'tcsh'
    BASH = 'bash'
    DASH = 'dash'
    FISH = 'fish'
    ZSH = 'zsh'
    CMD = 'cmd'


def subprocess_run(args, capture_output=True, text=False, **kwargs):
    # TODO: can upgrade arguments till runtime at Python 3.7
    #  subprocess.run(..., capture_output=True, text=True)
    stdout = subprocess.PIPE if capture_output else None
    stderr = subprocess.PIPE if capture_output else None
    process = subprocess.run(
        args, stdout=stdout, stderr=stderr, universal_newlines=text, **kwargs
    )
    return process


def get_usable_python3(python_command: str) -> Optional[str]:
    try:
        proc = subprocess_run([python_command, '-V'], text=True)
        output = str(proc.stderr) or str(proc.stdout)
        log.debug(f'stdout: {proc.stdout.strip()}, stderr: {proc.stderr.strip()}')
    except FileNotFoundError as e:
        output = ''
        log.exception(e)
    return python_command if 'Python 3' in output else None


def get_usable_lsf(lsf_command):
    try:
        proc = subprocess_run([lsf_command, '-V'], text=True)
        if str(proc.stderr):
            return lsf_command
    except FileNotFoundError:
        return None


def get_shell_output(shell_executoble=None, command=None) -> Optional[str]:
    try:
        command = command or 'echo $0'
        proc = subprocess_run(
            command, text=True, shell=True, executable=shell_executoble
        )
        stdout = proc.stdout.strip()
        return stdout if stdout != '$0' else None
    except FileNotFoundError:
        return None


def get_platform() -> Platform:
    system = platform.system()
    if 'Microsoft' in platform.release():
        return Platform.WSL
    elif get_usable_lsf('bsub'):
        return Platform.LSF
    else:
        return Platform(system)


def get_available_shell():
    avail_shells = set()

    if get_platform() == Platform.WINDOWS:
        avail_shells.add(Shell.CMD)
    if get_shell_output():
        avail_shells.add(Shell.DASH)
    for executable, test_command, shell in [
        ('/bin/bash', 'echo $BASH', Shell.BASH),
        ('/bin/tcsh', 'echo $shell', Shell.TCSH),
        ('/usr/bin/zsh', 'echo $ZSH_NAME', Shell.ZSH),
        ('/usr/bin/fish', 'echo $FISH_VERSION', Shell.FISH),
    ]:
        specific_shell_check = get_shell_output(executable, command=test_command)
        optional_shell_check = get_shell_output(executable)
        if executable == '/usr/bin/fish':
            optional_shell_check = True
        if specific_shell_check and optional_shell_check:
            avail_shells.add(shell)

    return avail_shells


PLATFORM = get_platform()
AVAIL_SHELLS = get_available_shell()
PYTHON = get_usable_python3('python3') or get_usable_python3('python') or sys.executable
BSUB = get_usable_lsf('bsub')


if __name__ == '__main__':
    print('Platform', PLATFORM)
    print('Avail shells', AVAIL_SHELLS)
    print('Python3', PYTHON)
    print('Bsub', BSUB)



import pathlib
import shutil
import uuid


def codepack(paths, output='output'):
    dest_folder = pathlib.Path(output)
    dest_folder.mkdir(exist_ok=True, parents=True)

    for path in paths:
        path = pathlib.Path(path)
        if path.is_dir():
            if path.is_absolute():
                sub_folder = path.stem
            else:
                rel_path_clipped = str(path).replace('..\\', '')
                sub_folder = pathlib.Path(rel_path_clipped)
            copy_folder(path, dest_folder / sub_folder)
        elif path.is_file():
            copy(path, dest_folder)


def generate_uuid(segment: int = 0):
    uid = uuid.uuid1().hex
    if segment:
        uid = uid[:segment]
    return uid


def copy_folder(src_path, dst_path):
    """Copy folder to specific destination.

    Args:
        src_path (str or :class:`pathlib.Path`): source file path.
        dst_path (str or :class:`pathlib.Path`): destination file path.
    """
    src_path = str(src_path)
    dst_path = str(dst_path)
    try:
        shutil.copytree(src_path, dst_path)
    except (shutil.Error, OSError) as e:
        print('Directory not copied. Error: %s' % e)


def copy(src_path, dst_path):
    """Copy file to specific destination.

    Args:
        src_path (str or :class:`pathlib.Path`): source file path.
        dst_path (str or :class:`pathlib.Path`): destination file path.
    """
    src_path = str(src_path)
    dst_path = str(dst_path)
    return shutil.copy(src_path, dst_path)



from yacs.config import CfgNode as _CfgNode


class CfgNode(_CfgNode):
    def merge_from_dict(self, dic):
        list_args = []
        for k, v in dic.items():
            list_args += [k, v]
        self.merge_from_list(list_args)


def create_config_node(path, new_allowed_levels=0):
    template = CfgNode(new_allowed=new_allowed_levels != 0)
    template.merge_from_file(path)

    for node in template.values():
        if isinstance(node, CfgNode):
            set_new_allow(node, new_allowed_levels - 1)
    return template


def set_new_allow(node, steps):
    node.__dict__[CfgNode.NEW_ALLOWED] = True
    if steps > 1:
        for sub_node in node.values():
            if isinstance(sub_node, CfgNode):
                set_new_allow(sub_node, steps - 1)



import logging


def console_handler(level=logging.WARNING):
    h = logging.StreamHandler()
    fmt = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    h.setLevel(level)
    h.setFormatter(fmt)
    return h


def file_handler(level=logging.WARNING, filename='klisch.log'):
    h = logging.FileHandler(filename)
    fmt = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    h.setLevel(level)
    h.setFormatter(fmt)
    return h


def get(name, level):
    log = logging.getLogger(name)
    log.addHandler(console_handler(level))
    log.setLevel(level)
    return log
